import asyncio
import datetime
import enum
import inspect
import uuid
import time
import json
import secrets
import itertools
import contextlib
from typing import Iterable, Coroutine


async def sleep(
    seconds: int, *, events: Iterable[Coroutine] | None = None
) -> bool:
    """
    Sleep for a specified number of seconds, or until one of the given events
    occurs.
    """
    if not events:
        await asyncio.sleep(seconds)
        return True

    tasks = [asyncio.create_task(event) for event in events]
    try:
        done, pending = await asyncio.wait(
            tasks,
            timeout=seconds,
            return_when=asyncio.FIRST_COMPLETED,
        )

        for task in pending:
            task.cancel()
    finally:
        for task in tasks:
            if not task.done():
                task.cancel()

    return True


@contextlib.contextmanager
def timed_block():
    """
    A context manager that times the block of code within it. The elapsed
    time will be updated in the `elapsed` attribute of the returned object
    until the block is exited.

    Example:

    .. code-block:: python

        with timed_block() as timer:
            time.sleep(1)

        print(f"Block took {timer.elapsed} seconds.")
    """

    class Timer:
        def __init__(self):
            self.start = time.monotonic()
            self.stop = None

        @property
        def elapsed(self):
            return (self.stop or time.monotonic()) - self.start

    timer = Timer()
    yield timer
    timer.stop = time.monotonic()


def importable_name(obj):
    """
    Get the importable name for an object.

    .. note::

        This will only work for objects that are actually importable,
        i.e. they are defined in a module. Lambdas, for example, will not
        have an importable name.

    :param obj: The object to get the importable name for.
    :return: str
    """
    if obj.__class__.__module__ == "__main__":
        module = inspect.getmodule(obj)
        return f"{module.__spec__.name}.{obj.__qualname__}"

    return f"{obj.__module__}.{obj.__qualname__}"


def import_string(name):
    """
    Import an object from a string previously created using
    :func:`importable_name`.

    :param name: The importable name of the object.
    :return: Any
    """
    mod_name, _, func_name = name.rpartition(".")
    mod = __import__(mod_name, fromlist=[func_name])
    return getattr(mod, func_name)


def chancy_uuid() -> str:
    """
    Generate a UUID suitable for use as a job ID.

    .. note::

        It's UUID7, kinda, since the draft keeps changing.

    :return: str
    """
    t = (time.time_ns() // 100) & 0xFFFFFFFFFFFFFF
    rand = secrets.randbits(62)
    uuid = (t << 68) | (7 << 64) | (2 << 62) | rand
    return f"{uuid:032x}"


def json_dumps(obj, **kwargs):
    """
    Serialize an object to a JSON formatted str with support for UUIDs.

    :param obj: The object to serialize.
    :param kwargs: Additional arguments to pass to `json.dumps`.
    :return: str
    """

    def _dump(o):
        if isinstance(o, uuid.UUID):
            return str(o)
        elif isinstance(o, datetime.datetime):
            return o.isoformat()
        elif isinstance(o, enum.Enum):
            return o.value

        raise TypeError(
            f"Object of type {o.__class__.__name__} is not JSON serializable"
        )

    return json.dumps(obj, default=_dump, **kwargs)


def chunked(iterable, size):
    """
    Yield chunks of `size` from `iterable`.

    :param iterable: The iterable to chunk.
    :param size: The size of each chunk.
    """
    it = iter(iterable)
    while True:
        chunk = list(itertools.islice(it, size))
        if not chunk:
            break
        yield chunk


class TaskManager:
    """
    A simple task manager that keeps track of tasks.
    """

    def __init__(self):
        self._tasks: set[asyncio.Task] = set()
        # An event triggered whenever add() is called to ensure that
        # wait_until_complete() will pick up new tasks.
        self._added_event = asyncio.Event()

    def add(self, task: Coroutine, *, name: str | None = None) -> asyncio.Task:
        """
        Add a task to the manager.
        """
        task = asyncio.create_task(task)
        task.add_done_callback(self._tasks.remove)

        if name:
            task.set_name(name)

        self._tasks.add(task)
        self._added_event.set()

        return task

    async def wait_until_complete(self):
        """
        Wait until all tasks are complete, including any that are added
        after this method is called.
        """
        while self._tasks:
            added_task = asyncio.create_task(
                self._added_event.wait(),
                name="TaskManager._added_event",
            )

            try:
                done, pending = await asyncio.wait(
                    self._tasks | {added_task},
                    return_when=asyncio.FIRST_COMPLETED,
                )

                if not added_task.done():
                    added_task.cancel()

                self._added_event.clear()

                for task in done - {added_task}:
                    if not task.cancelled():
                        task.result()
            finally:
                if not added_task.done():
                    added_task.cancel()

    async def cancel_all(self):
        """
        Cancel all tasks and wait for them to complete.
        """
        while self._tasks:
            for task in self._tasks:
                task.cancel()

            done, pending = await asyncio.wait(
                self._tasks,
                return_when=asyncio.FIRST_COMPLETED,
            )
            self._tasks = (pending | self._tasks) - done

    def __len__(self):
        return len(self._tasks)

    def __iter__(self):
        return iter(self._tasks)

    def __contains__(self, task: asyncio.Task):
        return task in self._tasks

    def __repr__(self):
        return f"<TaskManager tasks={len(self._tasks)!r}>"



"""
Utilities for creating dynamic rules that can be used when configuring the
conditions of a Plugin.
"""

from typing import Any
from psycopg import sql


class SQLAble:
    def to_sql(self) -> sql.Composable:
        raise NotImplementedError


class Rule(SQLAble):
    def __init__(self, field: str):
        self.field = field

    def __eq__(self, other: Any) -> "Condition":
        return Condition(self.to_sql(), "=", other)

    def __ne__(self, other: Any) -> "Condition":
        return Condition(self.to_sql(), "!=", other)

    def __lt__(self, other: Any) -> "Condition":
        return Condition(self.to_sql(), "<", other)

    def __le__(self, other: Any) -> "Condition":
        return Condition(self.to_sql(), "<=", other)

    def __gt__(self, other: Any) -> "Condition":
        return Condition(self.to_sql(), ">", other)

    def __ge__(self, other: Any) -> "Condition":
        return Condition(self.to_sql(), ">=", other)

    def __or__(self, other: "Condition") -> "OrCondition":
        return OrCondition(self, other)

    def __and__(self, other: "Condition") -> "AndCondition":
        return AndCondition(self, other)

    def to_sql(self) -> sql.Composable:
        return sql.Identifier(self.field)


class Condition(SQLAble):
    def __init__(self, field: sql.Composable, op: str, value: Any):
        self.field = field
        self.op = op
        self.value = value

    def __or__(self, other: "Condition") -> "OrCondition":
        return OrCondition(self, other)

    def __and__(self, other: "Condition") -> "AndCondition":
        return AndCondition(self, other)

    def to_sql(self) -> sql.Composable:
        return sql.SQL("{field} {op} {value}").format(
            field=self.field,
            op=sql.SQL(self.op),
            value=sql.Literal(self.value),
        )


class OrCondition(SQLAble):
    def __init__(self, left: SQLAble, right: SQLAble):
        self.left = left
        self.right = right

    def __or__(self, other: SQLAble) -> "OrCondition":
        return OrCondition(self, other)

    def __and__(self, other: SQLAble) -> "AndCondition":
        return AndCondition(self, other)

    def to_sql(self) -> sql.Composable:
        return sql.SQL("({left}) OR ({right})").format(
            left=self.left.to_sql(), right=self.right.to_sql()
        )


class AndCondition(SQLAble):
    def __init__(self, left: SQLAble, right: SQLAble):
        self.left = left
        self.right = right

    def __or__(self, other: Condition) -> OrCondition:
        return OrCondition(self, other)

    def __and__(self, other: Condition) -> "AndCondition":
        return AndCondition(self, other)

    def to_sql(self) -> sql.Composable:
        return sql.SQL("({left}) AND ({right})").format(
            left=self.left.to_sql(), right=self.right.to_sql()
        )


class JobRules:
    """
    A collection of rules that can be used to filter the main job table.
    """

    class Age(Rule):
        def __init__(self):
            super().__init__("age")

        def to_sql(self) -> sql.Composable:
            return sql.SQL("EXTRACT(EPOCH FROM (NOW() - created_at))")

    class Queue(Rule):
        def __init__(self):
            super().__init__("queue")

    class Job(Rule):
        def __init__(self):
            super().__init__("payload->>'func'")

    class State(Rule):
        def __init__(self):
            super().__init__("state")

    class CreatedAt(Rule):
        def __init__(self):
            super().__init__("created_at")

    class ScheduledAt(Rule):
        def __init__(self):
            super().__init__("scheduled_at")

    class ID(Rule):
        def __init__(self):
            super().__init__("id")



import click

from chancy import Chancy, Worker
from chancy.cli import run_async_command
from chancy.plugins.api import Api


@click.group(name="worker")
def worker_group():
    """
    Worker management commands.
    """
    pass


@worker_group.command("start")
@click.option("--worker-id", "-w", help="The worker ID to use.")
@click.option(
    "--tags",
    "-t",
    help="Extra tags to apply to the worker.",
    multiple=True,
)
@click.pass_context
@run_async_command
async def worker_command(
    ctx: click.Context, worker_id: str | None, tags: list[str] | None
):
    """
    Start a worker.
    """
    chancy: Chancy = ctx.obj["app"]

    async with chancy:
        async with Worker(
            chancy, worker_id=worker_id, tags=set(tags) if tags else None
        ) as worker:
            await worker.wait_until_complete()


@worker_group.command("web")
@click.option("--host", "-h", help="The host to bind to.", default="localhost")
@click.option("--port", "-p", help="The port to bind to.", default=8000)
@click.option(
    "--debug", "-d", help="Run the server in debug mode.", is_flag=True
)
@click.option(
    "--allow-origin",
    "-o",
    help="A list of allowed origins.",
    multiple=True,
    default=lambda: ["*"],
)
@click.pass_context
@run_async_command
async def web_command(
    ctx: click.Context,
    host: str,
    port: int,
    debug: bool,
    allow_origin: list[str],
):
    """
    Start the Chancy dashboard.
    """
    chancy: Chancy = ctx.obj["app"]

    async with chancy:
        api = Api(
            host=host,
            port=port,
            allow_origins=allow_origin,
            debug=debug,
        )

        worker = Worker(chancy)
        await api.run(worker, chancy)



import time
import asyncio

import pytest

from chancy import Chancy, Worker, Queue, Job, QueuedJob, Reference


def slow_job_to_run():
    time.sleep(5)


def job_to_run():
    pass


async def async_job_to_run():
    pass


def job_with_instance(*, job: QueuedJob):
    job.meta["received_instance"] = True


async def async_job_with_instance(*, job: QueuedJob):
    job.meta["received_instance"] = True


async def very_long_job():
    await asyncio.sleep(60 * 60)


def sync_very_long_job():
    time.sleep(60)


@pytest.mark.asyncio
async def test_basic_job_sync(
    chancy: Chancy, worker: tuple[Worker, asyncio.Task], sync_executor: str
):
    """
    Ensures that a basic job can be run on all built-in executors.
    """
    await chancy.declare(Queue("low", executor=sync_executor))

    ref = await chancy.push(Job.from_func(job_to_run, queue="low"))
    job = await chancy.wait_for_job(ref, timeout=30)

    assert job.state == job.State.SUCCEEDED


@pytest.mark.asyncio
async def test_basic_job_async(
    chancy: Chancy, worker: tuple[Worker, asyncio.Task], async_executor: str
):
    """
    Ensures that a basic job can be run on all built-in executors.
    """
    await chancy.declare(Queue("low", executor=async_executor))

    ref = await chancy.push(Job.from_func(async_job_to_run, queue="low"))
    job = await chancy.wait_for_job(ref, timeout=30)

    assert job.state == job.State.SUCCEEDED


@pytest.mark.asyncio
async def test_wait_for_job_timeout(
    chancy: Chancy, worker: tuple[Worker, asyncio.Task], sync_executor: str
):
    """
    Ensures that waiting for a job times out as expected.
    """
    await chancy.declare(Queue("low", executor=sync_executor))

    ref = await chancy.push(Job.from_func(job_to_run, queue="low"))
    job = await chancy.wait_for_job(ref, timeout=10)

    assert job.state == job.State.SUCCEEDED

    ref = await chancy.push(Job.from_func(slow_job_to_run, queue="low"))
    with pytest.raises(asyncio.TimeoutError):
        await chancy.wait_for_job(ref, timeout=0.1)


@pytest.mark.asyncio
async def test_job_instance_kwarg(
    chancy: Chancy, worker: tuple[Worker, asyncio.Task], sync_executor: str
):
    """
    Test that jobs requesting a QueuedJob kwarg receive the correct instance.
    """
    await chancy.declare(Queue("low", executor=sync_executor))

    ref = await chancy.push(Job.from_func(job_with_instance, queue="low"))
    job = await chancy.wait_for_job(ref, timeout=30)

    assert job.state == job.State.SUCCEEDED
    assert job.meta.get("received_instance") is True


@pytest.mark.asyncio
async def test_async_job_instance_kwarg(
    chancy: Chancy, worker: tuple[Worker, asyncio.Task], async_executor: str
):
    """
    Test that async jobs requesting a QueuedJob kwarg receive the correct
    instance.
    """
    await chancy.declare(Queue("low", executor=async_executor))

    ref = await chancy.push(Job.from_func(async_job_with_instance, queue="low"))
    job = await chancy.wait_for_job(ref, timeout=30)

    assert job.state == job.State.SUCCEEDED
    assert job.meta.get("received_instance") is True


@pytest.mark.asyncio
async def test_job_cancellation(
    chancy: Chancy, worker: tuple[Worker, asyncio.Task]
):
    """
    Test that jobs can be cancelled on supporting executors.
    """

    async def cancel_in_a_bit(to_cancel: Reference):
        await asyncio.sleep(10)
        await chancy.cancel_job(to_cancel)

    await chancy.declare(Queue("async", executor=Chancy.Executor.Async))
    await chancy.declare(Queue("sync", executor=Chancy.Executor.Process))

    ref = await chancy.push(Job.from_func(very_long_job, queue="async"))
    asyncio.create_task(cancel_in_a_bit(ref))
    job = await chancy.wait_for_job(ref, timeout=30)
    assert job.state == job.State.FAILED

    ref = await chancy.push(Job.from_func(sync_very_long_job, queue="sync"))
    asyncio.create_task(cancel_in_a_bit(ref))
    job = await chancy.wait_for_job(ref, timeout=30)
    assert job.state == job.State.FAILED



from psycopg import sql
from psycopg.rows import dict_row
from starlette.responses import Response

from chancy.plugins.api.plugin import ApiPlugin
from chancy.rule import JobRules
from chancy.utils import json_dumps


class CoreApiPlugin(ApiPlugin):
    """
    Core API plugin which implements the endpoints for jobs, queues, workers,
    etc...
    """

    def name(self):
        return "base"

    def routes(self):
        return [
            {
                "path": "/api/v1/configuration",
                "endpoint": self.get_configuration,
                "methods": ["GET"],
                "name": "get_configuration",
            },
            {
                "path": "/api/v1/queues",
                "endpoint": self.get_queues,
                "methods": ["GET"],
                "name": "get_queues",
            },
            {
                "path": "/api/v1/workers",
                "endpoint": self.get_workers,
                "methods": ["GET"],
                "name": "get_workers",
            },
            {
                "path": "/api/v1/jobs",
                "endpoint": self.get_jobs,
                "methods": ["GET"],
                "name": "get_jobs",
            },
            {
                "path": "/api/v1/jobs/{id}",
                "endpoint": self.get_job,
                "methods": ["GET"],
                "name": "get_job",
            },
        ]

    @staticmethod
    async def get_configuration(request, *, chancy, worker):
        """
        Get the configuration of the Chancy instance.
        """
        return Response(
            json_dumps(
                {
                    "plugins": [
                        plugin.__class__.__name__ for plugin in chancy.plugins
                    ]
                }
            ),
            media_type="application/json",
        )

    @staticmethod
    async def get_queues(request, *, chancy, worker):
        """
        Get a list of all the queues.
        """
        queues = await chancy.get_all_queues()
        return Response(
            json_dumps([queue.pack() for queue in queues]),
            media_type="application/json",
        )

    @staticmethod
    async def get_workers(request, *, chancy, worker):
        """
        Get a list of all the workers.
        """
        workers = await chancy.get_all_workers()
        return Response(
            json_dumps(workers),
            media_type="application/json",
        )

    @staticmethod
    async def get_jobs(request, *, chancy, worker):
        """
        Get a list of all the jobs in the system.

        Allows ID-based pagination and basic filtering.
        """
        state = request.query_params.get("state")
        queue = request.query_params.get("queue")
        limit = min(int(request.query_params.get("limit", 100)), 100)
        before = request.query_params.get("before")

        rule = JobRules.State() == "pending"

        if state:
            rule = JobRules.State() == state

        if before:
            rule &= JobRules.ID() < before

        if queue:
            rule &= JobRules.Queue() == queue

        async with chancy.pool.connection() as conn:
            async with conn.cursor(row_factory=dict_row) as cursor:
                await cursor.execute(
                    sql.SQL(
                        """
                        SELECT * FROM {jobs}
                        WHERE ({rule})
                        ORDER BY (
                            completed_at,
                            scheduled_at 
                        ) DESC
                        LIMIT {limit}
                        """
                    ).format(
                        jobs=sql.Identifier(f"{chancy.prefix}jobs"),
                        rule=rule.to_sql(),
                        limit=sql.Literal(limit),
                    )
                )

                result = await cursor.fetchall()

                if not result:
                    return Response(
                        json_dumps([]),
                        media_type="application/json",
                    )

                return Response(
                    json_dumps(result),
                    media_type="application/json",
                )

    @staticmethod
    async def get_job(request, *, chancy, worker):
        """
        Get a single job by ID.
        """
        job_id = request.path_params["id"]

        async with chancy.pool.connection() as conn:
            async with conn.cursor(row_factory=dict_row) as cursor:
                await cursor.execute(
                    sql.SQL(
                        """
                        SELECT * FROM {jobs}
                        WHERE id = %(job_id)s
                        """
                    ).format(
                        jobs=sql.Identifier(f"{chancy.prefix}jobs"),
                    ),
                    {"job_id": job_id},
                )

                result = await cursor.fetchone()

                if not result:
                    return Response(
                        json_dumps({}),
                        status_code=404,
                        media_type="application/json",
                    )

                return Response(
                    json_dumps(result),
                    media_type="application/json",
                )



"""
Executors
=========

This package contains the built-in executors for Chancy. Executors are
responsible for managing the execution of jobs after they've been retrieved
from a queue. A single worker can have multiple queues, each using its own
executor, allowing you to run jobs in the way that best suits their needs.

Capabilities
------------

Not all executors have the same capabilities, typically due to limits in how
threads are implemented in Python. The following table summarizes the
capabilities of each executor:

.. list-table:: Executor Capabilities
   :header-rows: 1
   :widths: 20 15 15 15

   * - Executor
     - Cancellation [#f1]_
     - Timeouts
     - Memory Limits
   * - ProcessExecutor
     - ✓
     - ✓
     - ✓
   * - AsyncExecutor
     - ✓
     - ✓
     - ✗
   * - ThreadedExecutor
     - ✗
     - ✓
     - ✗
   * - SubInterpreter
     - ✗
     - ✓
     - ✗

.. [#f1] Cancellation is always possible before a job is started. Cancellation
         here refers to the ability to stop a job that is actively running.

ProcessExecutor (Default)
~~~~~~~~~~~~~~~~~~~~~~~~~
:class:`chancy.executors.process.ProcessExecutor`
Optimized for CPU-bound jobs, uses separate processes for true parallelism.

AsyncExecutor
~~~~~~~~~~~~~
:class:`chancy.executors.asyncex.AsyncExecutor`
Optimized for IO-bound jobs like API calls and database operations. Uses
asyncio for efficient concurrent execution. Thousands of jobs can be
executed concurrently with limited resources.

ThreadedExecutor
~~~~~~~~~~~~~~~~
:class:`chancy.executors.thread.ThreadedExecutor`
Suitable for IO-bound jobs that can't use asyncio. Uses threads for concurrent
execution.

SubInterpreterExecutor
~~~~~~~~~~~~~~~~~~~~~~
:class:`chancy.executors.sub.SubInterpreterExecutor`
Experimental executor using Python sub-interpreters. Provides GIL avoidance
with lower overhead than processes.


Custom Executors
~~~~~~~~~~~~~~~~
You can implement your own executor by subclassing the
:class:`chancy.executors.base.Executor` class and implementing the
:meth:`~chancy.executors.base.Executor.push` method.
"""



import asyncio

import pytest

from chancy.app import Chancy
from chancy.worker import Worker
from chancy.plugins.leadership import ImmediateLeadership, Leadership


@pytest.mark.parametrize(
    "chancy",
    [
        {
            "plugins": [
                ImmediateLeadership(),
            ]
        }
    ],
    indirect=True,
)
@pytest.mark.asyncio
async def test_immediate_leadership(chancy: Chancy, worker: Worker):
    """
    Ensures that the immediate leadership plugin works as expected.
    """
    assert worker.is_leader.is_set()


@pytest.mark.parametrize(
    "chancy",
    [
        {
            "plugins": [
                Leadership(poll_interval=5),
            ]
        }
    ],
    indirect=True,
)
@pytest.mark.asyncio
async def test_leadership(chancy: Chancy, worker: Worker):
    """
    Ensures that the leadership plugin works as expected.
    """
    await asyncio.sleep(10)
    assert worker.is_leader.is_set()


@pytest.mark.parametrize(
    "chancy",
    [
        {
            "plugins": [
                Leadership(poll_interval=5, timeout=10),
            ]
        }
    ],
    indirect=True,
)
@pytest.mark.asyncio
async def test_leadership_many_workers(chancy: Chancy):
    """
    Ensures that the leadership plugin works as expected with several workers.
    """
    workers = [Worker(chancy) for _ in range(10)]
    for worker in workers:
        await worker.start()

    await asyncio.sleep(15)
    leaders = [worker for worker in workers if worker.is_leader.is_set()]
    assert len(leaders) == 1

    for worker in workers:
        await worker.stop()


@pytest.mark.parametrize(
    "chancy",
    [
        {
            "plugins": [
                Leadership(poll_interval=5, timeout=10),
            ]
        }
    ],
    indirect=True,
)
@pytest.mark.asyncio
async def test_leadership_transition(chancy: Chancy):
    """
    Ensures that leadership transitions work as expected.
    """
    worker1 = Worker(chancy, worker_id="1")
    worker2 = Worker(chancy, worker_id="2")

    await worker1.start()
    await worker2.start()

    await asyncio.sleep(15)

    # Only one worker should have become the leader.
    assert worker1.is_leader.is_set() != worker2.is_leader.is_set()

    # Who got the leadership.
    who_leads = worker1 if worker1.is_leader.is_set() else worker2
    who_follows = worker2 if worker1.is_leader.is_set() else worker1

    await who_leads.stop()

    await asyncio.sleep(15)

    assert who_follows.is_leader.is_set()
    await who_follows.stop()



import asyncio
import threading
import functools
from concurrent.futures import Future

import sys

from chancy.worker import Worker
from chancy.queue import Queue

try:
    # Only available in 3.14+
    from concurrent.futures import InterpreterPoolExecutor
except ImportError:
    # Available only in 3.13 with a backport.
    try:
        from interpreters_backport.concurrent.futures.interpreter import (
            InterpreterPoolExecutor,
        )
    except ImportError:
        raise ImportError(
            "The SubInterpreterExecutor requires Python 3.13 or later."
        )

from chancy.executors.base import Executor, ConcurrentExecutor
from chancy.job import QueuedJob, Limit


class SubInterpreterExecutor(ConcurrentExecutor):
    """
    .. note::

        This executor is experimental and may not work as expected. The
        sub-interpreter features it is built on is experimental and only
        available in Python 3.13 (with a backport) and >3.14.

        To use with 3.13, run:

        .. code-block:: bash

            pip install chancy[sub]

    To use this executor, simply pass the import path to this class in the
    ``executor`` field of your queue configuration or use the
    :class:`~chancy.app.Chancy.Executor` shortcut:

    .. code-block:: python

        async with Chancy(dsn="postgresql://localhost/postgres") as chancy:
            await chancy.declare(
                Queue(
                    name="default",
                    executor=Chancy.Executor.SubInterpreter,
                )
            )

    It's important to note that many C-based Python libraries are not (yet)
    compatible with sub-interpreters. If you encounter issues, you may need to
    switch to a different executor.

    :param worker: The worker instance associated with this executor.
    :param queue: The queue that this executor is associated with.
    """

    def __init__(self, worker: Worker, queue: Queue):
        super().__init__(worker, queue)
        self.pool = InterpreterPoolExecutor(
            max_workers=queue.concurrency,
            initializer=self.on_initialize_worker,
            initargs=(sys.path,),
        )

    @staticmethod
    def on_initialize_worker(parent_sys_path: list[str]):
        """
        This method is called in each worker before it begins running jobs.
        It can be used to perform any necessary setup, such as loading NLTK
        datasets or calling ``django.setup()``.

        By default, it replaces the running job's ``sys.path`` with the workers.
        """
        # Unlike all other executors, the InterpreterPoolExecutor does not
        # automatically inherit the parent process's sys.path. This is a
        # workaround to ensure that the worker has the same sys.path as the
        # parent process or tests will fail.
        sys.path = parent_sys_path

    async def push(self, job: QueuedJob) -> Future:
        future: Future = self.pool.submit(self.job_wrapper, job)
        self.jobs[future] = job
        future.add_done_callback(
            functools.partial(
                self._on_job_completed, loop=asyncio.get_running_loop()
            )
        )
        return future

    @staticmethod
    def job_wrapper(job: QueuedJob):
        """
        This is the function that is actually started by the sub-interpreter
        executor. It's responsible for setting up necessary limits,
        running the job, and returning the result.
        """
        try:
            time_limit = next(
                (
                    limit.value
                    for limit in job.limits
                    if limit.type_ == Limit.Type.TIME
                ),
                None,
            )

            func, kwargs = Executor.get_function_and_kwargs(job)

            if time_limit:
                timer = threading.Timer(
                    time_limit, SubInterpreterExecutor._timeout_handler
                )
                timer.start()
                try:
                    func(**kwargs)
                finally:
                    timer.cancel()
            else:
                func(**kwargs)

        except Exception as exc:
            return exc
        return job

    @staticmethod
    def _timeout_handler():
        raise TimeoutError("Job timed out.")

    def _on_job_completed(
        self, future: Future, loop: asyncio.AbstractEventLoop
    ):
        job = self.jobs.pop(future)
        exc = future.exception()
        if exc is None:
            job = future.result()

        f = asyncio.run_coroutine_threadsafe(
            self.job_completed(job, exc),
            loop,
        )
        f.result()

    async def stop(self):
        self.pool.shutdown(cancel_futures=True)



import threading
from concurrent.futures import ThreadPoolExecutor, Future
import asyncio
import functools

from chancy.executors.base import Executor, ConcurrentExecutor
from chancy.job import QueuedJob, Limit


class ThreadedExecutor(ConcurrentExecutor):
    """
    An Executor which uses a thread pool to run its jobs.

    This executor is useful for running I/O-bound jobs concurrently without the
    overhead of separate processes. It's not suitable for CPU-bound tasks due
    to Python's Global SubInterpreter Lock (GIL).

    When working with existing asyncio code, it's often easier and more
    efficient to use the :class:`~chancy.executors.asyncex.AsyncExecutor`
    instead, as it can run a very large number of jobs concurrently.

    To use this executor, simply pass the import path to this class in the
    ``executor`` field of your queue configuration or use the
    :class:`~chancy.app.Chancy.Executor` shortcut:

    .. code-block:: python

        async with Chancy(dsn="postgresql://localhost/postgres") as chancy:
            await chancy.declare(
                Queue(
                    name="default",
                    executor=Chancy.Executor.Threaded,
                )
            )

    :param worker: The worker instance associated with this executor.
    :param queue: The queue that this executor is associated with.
    """

    def __init__(self, worker, queue):
        super().__init__(worker, queue)
        self.pool = ThreadPoolExecutor(max_workers=queue.concurrency)

    async def push(self, job: QueuedJob) -> Future:
        future: Future = self.pool.submit(self.job_wrapper, job)
        self.jobs[future] = job
        future.add_done_callback(
            functools.partial(
                self._on_job_completed, loop=asyncio.get_running_loop()
            )
        )
        return future

    def job_wrapper(self, job: QueuedJob):
        """
        This is the function that is actually started by the thread pool
        executor. It's responsible for setting up necessary limits,
        running the job, and returning the result.
        """
        func, kwargs = Executor.get_function_and_kwargs(job)
        if asyncio.iscoroutinefunction(func):
            raise ValueError(
                f"Function {job.func!r} is an async function, which is not"
                f" supported by the {self.__class__.__name__!r} executor. Use"
                f" the AsyncExecutor instead."
            )

        time_limit = next(
            (
                limit.value
                for limit in job.limits
                if limit.type_ == Limit.Type.TIME
            ),
            None,
        )

        if time_limit:
            timer = threading.Timer(time_limit, self._timeout_handler)
            timer.start()
            try:
                func(**kwargs)
            finally:
                timer.cancel()
        else:
            func(**kwargs)

        return job

    @staticmethod
    def _timeout_handler():
        raise TimeoutError("Job timed out.")

    def _on_job_completed(
        self, future: Future, loop: asyncio.AbstractEventLoop
    ):
        job = self.jobs.pop(future)

        exc = future.exception()
        if exc is None:
            job = future.result()

        asyncio.run_coroutine_threadsafe(
            self.job_completed(job, exc),
            loop,
        )

    async def stop(self):
        self.pool.shutdown(cancel_futures=True)
        await super().stop()



from psycopg import sql

from chancy.plugin import Plugin, PluginScope
from chancy.rule import JobRules


class Reprioritize(Plugin):
    """
    A plugin that increases the priority of jobs based on how long they've been
    in the queue. This helps prevent job starvation by gradually increasing the
    priority of older jobs.

    Example usage:

    .. code-block:: python

        plugin = Reprioritize(
            rule=(
                (Reprioritize.Rules.Age() > 600) &
                (Reprioritize.Rules.Queue() == "high-priority")
            ),
            check_interval=300,
            priority_increase=1
        )

    This means that any job that has been in the queue for more than 10 minutes
    will have its priority increased by 1 every 5 minutes, but only for jobs
    in the high-priority queue.
    """

    Rules = JobRules

    def __init__(
        self,
        rule,
        *,
        check_interval: int = 300,
        priority_increase: int = 1,
        batch_size: int = 1000,
    ):
        super().__init__()
        self.rule = rule
        self.check_interval = check_interval
        self.priority_increase = priority_increase
        self.batch_size = batch_size

    @classmethod
    def get_scope(cls) -> PluginScope:
        return PluginScope.WORKER

    async def run(self, worker, chancy):
        while await self.sleep(self.check_interval):
            await self.wait_for_leader(worker)
            updated = await self.reprioritize_jobs(chancy)
            chancy.log.info(
                f"Reprioritized {updated} jobs by increasing priority"
                f" by {self.priority_increase}"
            )

    async def reprioritize_jobs(self, chancy) -> int:
        """
        Reprioritize jobs based on the rules provided to the plugin.

        Returns the total number of jobs that were updated.
        """
        total_updated = 0

        async with chancy.pool.connection() as conn:
            async with conn.cursor() as cursor:
                while True:
                    # Update jobs in batches to avoid long-running transactions
                    async with conn.transaction():
                        await cursor.execute(
                            sql.SQL(
                                """
                                WITH jobs_to_update AS (
                                    SELECT 
                                        id,
                                        priority
                                    FROM {jobs_table}
                                    WHERE 
                                        state IN ('pending', 'retrying')
                                        AND ({rule})
                                    ORDER BY id
                                    LIMIT {batch_size}
                                    FOR UPDATE SKIP LOCKED
                                )
                                UPDATE {jobs_table} j
                                SET 
                                    priority = j.priority + {increment}
                                FROM jobs_to_update
                                WHERE j.id = jobs_to_update.id
                                RETURNING j.id
                                """
                            ).format(
                                jobs_table=sql.Identifier(
                                    f"{chancy.prefix}jobs"
                                ),
                                rule=self.rule.to_sql(),
                                increment=self.priority_increase,
                                batch_size=self.batch_size,
                            )
                        )

                        results = await cursor.fetchall()
                        if not results:
                            break

                        total_updated += len(results)

        return total_updated



import abc
import inspect
import typing
import traceback
import dataclasses
from abc import ABC
from asyncio import Future
from datetime import datetime, timezone
from typing import get_type_hints, Callable

from chancy.job import QueuedJob, Reference
from chancy.queue import Queue

if typing.TYPE_CHECKING:
    from chancy.worker import Worker


class Executor(abc.ABC):
    """
    The base class for all executors.

    Executors are responsible for managing the execution of jobs after they've
    been retrieved from a queue.

    See :class:`~chancy.executors.process.ProcessExecutor` and
    :class:`~chancy.executors.asyncex.AsyncExecutor` for examples of built-in
    executors.
    """

    def __init__(self, worker: "Worker", queue: "Queue"):
        self.worker = worker
        self.queue = queue

    @abc.abstractmethod
    async def push(self, job: QueuedJob):
        """
        Push a job onto the job pool.
        """

    async def job_completed(self, job: QueuedJob, exc: Exception | None = None):
        """
        Called when a job has completed.

        This method should be called by the executor when a job has completed
        execution. It will update the job's state in the queue and handle
        retries if necessary.

        :param job: The job that has completed.
        :param exc: The exception that was raised during execution, if any.
        """
        if exc is None:
            now = datetime.now(tz=timezone.utc)
            new_instance = dataclasses.replace(
                job,
                state=QueuedJob.State.SUCCEEDED,
                completed_at=now,
                attempts=job.attempts + 1,
            )
        else:
            is_failure = job.attempts + 1 >= job.max_attempts

            new_state = (
                QueuedJob.State.FAILED
                if is_failure
                else QueuedJob.State.RETRYING
            )

            new_instance = dataclasses.replace(
                job,
                state=new_state,
                attempts=job.attempts + 1,
                completed_at=(
                    datetime.now(tz=timezone.utc) if is_failure else None
                ),
                errors=[
                    *job.errors,
                    {
                        "traceback": "".join(
                            traceback.format_exception(
                                type(exc), exc, exc.__traceback__
                            )
                        ),
                        "attempt": job.attempts,
                    },
                ],
            )

        # Each plugin has a chance to modify the job instance after it's
        # completed.
        for plugin in self.worker.chancy.plugins:
            try:
                new_instance = await plugin.on_job_completed(
                    new_instance, worker=self.worker, exc=exc
                )
            except NotImplementedError:
                continue

        await self.worker.queue_update(new_instance)

    @staticmethod
    def get_function_and_kwargs(job: QueuedJob) -> tuple[Callable, dict]:
        """
        Finds the function which should be executed for the given job and
        returns its keyword arguments.

        :param job: The job instance to get the function and arguments for.
        :return: A tuple containing the function and its keyword arguments.
        """
        mod_name, func_name = job.func.rsplit(".", 1)
        mod = __import__(mod_name, fromlist=[func_name])
        try:
            func = getattr(mod, func_name)
        except AttributeError:
            raise AttributeError(
                f"Could not find function {func_name} in module {mod_name}."
            )

        # We take a look at the type signature for the function to see if the
        # user has specified that the job instance should be passed as a
        # keyword argument.
        type_hints = get_type_hints(func)
        sig = inspect.signature(func)
        kwargs = job.kwargs or {}
        for param_name, param in sig.parameters.items():
            if not param.kind == param.KEYWORD_ONLY:
                continue

            type_hint = type_hints.get(param_name)
            if type_hint is None:
                continue

            if issubclass(type_hint, QueuedJob):
                kwargs[param_name] = job

        return func, kwargs

    @abc.abstractmethod
    async def stop(self):
        """
        Stop the executor, giving it a chance to clean up any resources it
        may have allocated to running jobs.

        It is not safe to use the executor after this method has been
        called.
        """

    @abc.abstractmethod
    async def cancel(self, ref: Reference):
        """
        Attempt to cancel a running job.

        It's not guaranteed that the job will be cancelled, as it may have
        already completed by the time this method is called or the executor
        may not support cancellation.

        :param ref: The reference to the job to cancel.
        """

    @abc.abstractmethod
    def __len__(self):
        """
        Get the number of jobs currently within the executor.
        """

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc, tb):
        await self.stop()


class ConcurrentExecutor(Executor, ABC):
    """
    Base class for executors based off of the concurrent.futures.Executor
    class with common functionality.
    """

    def __init__(self, worker: "Worker", queue: "Queue"):
        super().__init__(worker, queue)
        self.jobs: dict[Future, QueuedJob] = {}

    async def cancel(self, ref: Reference):
        for future, job in self.jobs.items():
            if job.id == ref.identifier:
                future.cancel()
                return

    def __len__(self):
        return len(self.jobs)



"""
Jobs
====

In Chancy, a Job represents a unit of work to be executed asynchronously.
Understanding how to create, configure, and manage jobs is crucial for
effectively using Chancy.

Creating a Job
--------------

There are two main ways to create a job:

1. From a Python function:

   .. code-block:: python

      from chancy import Job

      def greet(name: str):
          print(f"Hello, {name}!")

      job = Job.from_func(greet, kwargs={"name": "Alice"})

2. From a string import path:

   .. code-block:: python

      job = Job(func="mymodule.greet", kwargs={"name": "Bob"})


Jobs are immutable once created - use the `with_` methods on a Job to create
a new job with modified properties.


Priority
^^^^^^^^

Priority determines the order of execution. The higher the priority, the
sooner the job will be executed:

.. code-block:: python

   high_priority_job = job.with_priority(10)
   low_priority_job = job.with_priority(-10)

Retry Attempts
^^^^^^^^^^^^^^

Specify how many times a job should be retried if it fails:

.. code-block:: python

   Job.from_func(greet, kwargs={"name": "Charlie"}, max_attempts=3)

Scheduled Execution
^^^^^^^^^^^^^^^^^^^

Schedule a job to run some time in the future:

.. code-block:: python

   from datetime import datetime, timedelta, timezone

   future_job = Job.from_func(greet, kwargs={"name": "David"})
   future_job = future_job.with_scheduled_at(
       datetime.now(timezone.utc) + timedelta(hours=1)
   )

.. note::

    Scheduled jobs are guaranteed to run *at* or *after* the scheduled time,
    but not *exactly* at that time.

If you need recurring jobs, take a look at the
:class:`~chancy.plugins.cron.Cron` plugin.

Resource Limits
^^^^^^^^^^^^^^^

Set memory and time limits for job execution:

.. code-block:: python

   from chancy import Limit

   limited_job = Job.from_func(
       greet,
       kwargs={"name": "Eve"},
       limits=[
           Limit(Limit.Type.MEMORY, 1024 * 1024 * 1024),  # 1GB
           Limit(Limit.Type.TIME, 60),  # 60 seconds
       ]
   )

Queueing a Job
--------------

Once you've created a job, queue it for execution:

.. code-block:: python

   async with Chancy(dsn="postgresql://localhost/postgres") as chancy:
       await chancy.push(job)

Queue multiple jobs at once:

.. code-block:: python

   await chancy.push_many([job1, job2, job3])

If you want to queue jobs within a transaction, use the
:meth:`~chancy.app.Chancy.push_many_ex` method.

Push returns a :class:`~chancy.job.Reference` object that can be used to
retrieve the job instance later, or wait for it to complete.

Unique Jobs
-----------

Prevent duplicate job execution by assigning a unique key:

.. code-block:: python

   user_id = 1234
   unique_job = Job(
      func="generate_report",
      kwargs={"user_id": user_id},
      unique_key=f"report_{user_id}"
   )

.. note::

  Unique jobs ensure only one job with the same ``unique_key`` is
  queued or running at a time, but any number can be completed or
  failed.
"""

import dataclasses
import enum
from datetime import datetime, timezone
from typing import Any, Optional, TypedDict

from chancy.utils import importable_name


class ErrorT(TypedDict):
    #: The attempt number of the job when the error occurred.
    attempt: int
    #: The error message, typically the traceback of an exception.
    traceback: str


class Reference:
    """
    References a Job in the queue.

    This object can be used to retrieve the job instance later, or wait for it
    to complete. It is returned by the :meth:`~chancy.app.Chancy.push`,
    :meth:`~chancy.app.Chancy.push_many`, and
    :meth:`~chancy.app.Chancy.push_many_ex` functions.

    Waiting for a job to finish:

    .. code-block:: python

         async with Chancy(dsn="postgresql://localhost/postgres") as chancy:
            ref = await chancy.push(Job.from_func(my_function))
            job = await chancy.wait_for_job(ref)
            print(job.state)  # "succeeded"

    Retrieving a job instance:

    .. code-block:: python

        async with Chancy(dsn="postgresql://localhost/postgres") as chancy:
            ref = await chancy.push(Job.from_func(my_function))
            job = await chancy.get_job(ref)
            print(job.state)  # "pending"
    """

    __slots__ = ("identifier",)

    def __init__(self, identifier: str):
        self.identifier = identifier

    def __repr__(self):
        return f"<Reference({self.identifier!r})>"


@dataclasses.dataclass
class Limit:
    """
    A limit that can be applied to a job.
    """

    class Type(enum.Enum):
        MEMORY = "memory"
        TIME = "time"

    type_: "Type"
    value: int

    @classmethod
    def deserialize(cls, data: dict) -> "Limit":
        return cls(type_=cls.Type(data["t"]), value=data["v"])

    def serialize(self) -> dict:
        return {"t": self.type_.value, "v": self.value}


@dataclasses.dataclass(frozen=True, kw_only=True)
class Job:
    """
    A job is an immutable, stateless unit of work that can be pushed onto a
    Chancy queue and executed elsewhere.
    """

    #: An importable name for the function that should be executed when this
    #: job is run. Ex: my_module.my_function
    func: str
    #: The queue to which this job should be pushed.
    queue: str = "default"
    #: The keyword arguments to pass to the job function when it is executed.
    kwargs: dict[str, Any] | None = dataclasses.field(default_factory=dict)
    #: The priority of this job. Jobs with higher priority values will be
    #: executed before jobs with lower priority values.
    priority: int = 0
    #: The maximum number of times this job can be attempted before it is
    #: considered failed.
    max_attempts: int = 1
    #: The time at which this job should be scheduled to run.
    scheduled_at: datetime = dataclasses.field(
        default_factory=lambda: datetime.now(tz=timezone.utc)
    )
    #: A list of resource limits that should be applied to this job.
    limits: list[Limit] = dataclasses.field(default_factory=list)
    #: An optional, globally unique identifier for this job. If provided,
    #: only 1 copy of a job with this key will be allowed to run or be
    #: scheduled at a time.
    unique_key: str | None = None
    #: Arbitrary metadata associated with this job instance. Plugins can use
    #: this to store additional information during the execution of a job.
    meta: dict[str, Any] = dataclasses.field(default_factory=dict)

    @classmethod
    def from_func(cls, func, **kwargs):
        return cls(func=importable_name(func), **kwargs)

    def with_priority(self, priority: int) -> "Job":
        return dataclasses.replace(self, priority=priority)

    def with_max_attempts(self, max_attempts: int) -> "Job":
        return dataclasses.replace(self, max_attempts=max_attempts)

    def with_scheduled_at(self, scheduled_at: datetime) -> "Job":
        return dataclasses.replace(self, scheduled_at=scheduled_at)

    def with_limits(self, limits: list[Limit]) -> "Job":
        return dataclasses.replace(self, limits=limits)

    def with_kwargs(self, kwargs: dict[str, Any]) -> "Job":
        return dataclasses.replace(self, kwargs=kwargs)

    def with_unique_key(self, unique_key: str) -> "Job":
        return dataclasses.replace(self, unique_key=unique_key)

    def with_queue(self, queue: str) -> "Job":
        return dataclasses.replace(self, queue=queue)

    def with_meta(self, meta: dict[str, Any]) -> "Job":
        return dataclasses.replace(self, meta=meta)

    def pack(self) -> dict:
        """
        Pack the job into a dictionary that can be serialized and used to
        recreate the job later.
        """
        return {
            "f": self.func,
            "k": self.kwargs,
            "p": self.priority,
            "a": self.max_attempts,
            "s": self.scheduled_at.timestamp(),
            "l": [limit.serialize() for limit in self.limits],
            "u": self.unique_key,
            "q": self.queue,
            "m": self.meta,
        }

    @classmethod
    def unpack(cls, data: dict) -> "Job":
        """
        Unpack a serialized job into a Job instance.
        """
        return cls(
            func=data["f"],
            kwargs=data["k"],
            priority=data["p"],
            max_attempts=data["a"],
            scheduled_at=datetime.fromtimestamp(data["s"], tz=timezone.utc),
            limits=[Limit.deserialize(limit) for limit in data["l"]],
            unique_key=data["u"],
            queue=data["q"],
            meta=data["m"],
        )


@dataclasses.dataclass(frozen=True, kw_only=True)
class QueuedJob(Job):
    """
    A job instance is a job that has been pushed onto a queue and now has
    stateful information associated with it, such as the number of attempts
    so far.
    """

    class State(enum.Enum):
        PENDING = "pending"
        RUNNING = "running"
        FAILED = "failed"
        RETRYING = "retrying"
        SUCCEEDED = "succeeded"

    #: The unique identifier for this job instance.
    id: str
    #: The time at which this job was started, if it has been started.
    started_at: Optional[datetime] = None
    #: The time at which this job was completed, if it has been completed.
    completed_at: Optional[datetime] = None
    #: The number of times this job has been attempted.
    attempts: int = 0
    #: The current state of this job instance.
    state: State = State.PENDING
    #: A list of errors that occurred during the execution of this job.
    errors: list[ErrorT] = dataclasses.field(default_factory=list)

    @classmethod
    def unpack(cls, data: dict) -> "QueuedJob":
        return cls(
            id=str(data["id"]),
            func=data["func"],
            kwargs=data["kwargs"],
            priority=data["priority"],
            scheduled_at=data["scheduled_at"],
            started_at=data["started_at"],
            completed_at=data["completed_at"],
            attempts=data["attempts"],
            max_attempts=data["max_attempts"],
            state=QueuedJob.State(data["state"]),
            unique_key=data["unique_key"],
            queue=data["queue"],
            errors=data["errors"],
            limits=[Limit.deserialize(limit) for limit in data["limits"]],
            meta=data["meta"],
        )



import os
from pathlib import Path
from functools import partial
from typing import Type

import uvicorn
from starlette.applications import Starlette
from starlette.middleware import Middleware
from starlette.middleware.cors import CORSMiddleware
from starlette.staticfiles import StaticFiles

from chancy import Worker, Chancy
from chancy.plugin import Plugin, PluginScope
from chancy.plugins.api.core import CoreApiPlugin
from chancy.plugins.api.plugin import ApiPlugin
from chancy.utils import import_string


class _SPAStaticFiles(StaticFiles):
    """
    A StaticFiles class that serves the SPA index.html file for any path that
    doesn't match an existing file.
    """

    def lookup_path(self, path: str) -> tuple[str, os.stat_result | None]:
        full_path, stat_result = super().lookup_path(path)
        if stat_result is None:
            return super().lookup_path("./index.html")
        return full_path, stat_result


class Api(Plugin):
    """
    Provides an API and a dashboard for viewing the state of the Chancy cluster.

    Running this plugin requires a few additional dependencies, you can install
    them with:

    .. code-block:: bash

        pip install chancy[web]

    Then add the API plugin to your Chancy instance:

    .. code-block:: python

        from chancy.plugins.api import Api

        async with Chancy(..., plugins=[
            WorkflowPlugin(),
            Api(),
        ]) as chancy:
            ...

    .. note::

        The API is still mostly undocumented, as its development is driven by
        the needs of the dashboard and may change significantly before it
        becomes stable.

    .. warning::

        The api interface is not secure and should not be exposed to the
        public internet. It is intended for use in a secure environment, such
        as a private network or a VPN where only trusted users have access.

    :param port: The port to listen on.
    :param host: The host to listen on.
    :param debug: Whether to run the server in debug mode.
    :param allow_origins: A list of origins that are allowed to access the API.
    """

    @classmethod
    def get_scope(cls) -> PluginScope:
        return PluginScope.WORKER

    def __init__(
        self,
        *,
        port: int = 8000,
        host: str = "127.0.0.1",
        debug: bool = False,
        allow_origins: list[str] | None = None,
    ):
        super().__init__()
        self.port = port
        self.host = host
        self.debug = debug
        self.root = Path(__file__).parent
        self.allow_origins = allow_origins or []
        self.plugins: set[Type[ApiPlugin]] = {CoreApiPlugin}

    async def run(self, worker: Worker, chancy: Chancy):
        """
        Start the web server.
        """
        for plug in chancy.plugins:
            api_plugin = plug.api_plugin()
            if api_plugin is None:
                continue

            api_plugin = import_string(api_plugin)
            if not issubclass(api_plugin, ApiPlugin):
                continue

            self.plugins.add(api_plugin)

        def _r(f):
            return partial(f, chancy=chancy, worker=worker)

        app = Starlette(
            debug=self.debug,
            middleware=[
                Middleware(
                    CORSMiddleware,
                    allow_origins=self.allow_origins,
                )
            ],
        )

        web_plugins = []
        # Look through all the enabled plugins for any that implement the
        # ApiPlugin interface. If they do, we merge them into our Starlette
        # app.
        for api_plugin in self.plugins:
            wp = api_plugin()
            web_plugins.append(wp)
            chancy.log.info(f"Loading API sub-plugin {wp.name()}")

            for route in wp.routes():
                app.add_route(
                    route["path"],
                    _r(route["endpoint"]),
                    methods=route["methods"],
                    name=route["name"],
                )

        # Add the wildcard route to the end of the list so that it doesn't
        # override any other routes and serves anything that should be handled
        # by the UI SPA.
        app.mount(
            "/",
            _SPAStaticFiles(
                packages=[("chancy.plugins.api", "dist")],
                html=True,
            ),
            name="ui",
        )

        server = uvicorn.Server(
            config=uvicorn.Config(
                app=app,
                host=self.host,
                port=self.port,
                log_level="error",
            )
        )

        chancy.log.info(f"Dashboard running at http://{self.host}:{self.port}")

        await server.serve()



import asyncio
import functools
import multiprocessing
import os
from multiprocessing.context import BaseContext

import resource
import signal
from asyncio import Future, CancelledError
from concurrent.futures import ProcessPoolExecutor
from typing import Callable

from chancy import Reference
from chancy.executors.base import ConcurrentExecutor
from chancy.job import QueuedJob, Limit


class ProcessExecutor(ConcurrentExecutor):
    """
    An Executor which uses a process pool to run its jobs.

    This executor is useful for running jobs that are CPU-bound, avoiding the
    GIL (Global SubInterpreter Lock) that Python uses to ensure thread safety.

    To use this executor, simply pass the import path to this class in the
    ``executor`` field of your queue configuration or use the
    :class:`~chancy.app.Chancy.Executor` shortcut:

    .. code-block:: python

        async with Chancy(dsn="postgresql://localhost/postgres") as chancy:
            await chancy.declare(
                Queue(
                    name="default",
                    concurrency=10,
                    executor=Chancy.Executor.Process,
                    executor_options={
                        "maximum_jobs_per_worker": 100,
                    }
                )
            )


    :param queue: The queue that this executor is associated with.
    :param maximum_jobs_per_worker: The maximum number of jobs that each worker
                                    can run before being replaced. Handy if you
                                    are stuck using a library with memory leaks.
    :param mp_context: The multiprocessing context to use. If not provided, the
                       default "spawn" context will be used, which is the
                       safest option on all platforms.
    """

    def __init__(
        self,
        worker,
        queue,
        *,
        maximum_jobs_per_worker: int = 100,
        mp_context: BaseContext | None = None,
    ):
        super().__init__(worker, queue)
        # Becomes the default on Linux as of 3.14 due to potential crashes
        # in the `fork` method if the child process also works with threads.
        # We're using `spawn` explicitly to get ahead of the curve, however
        # this is slower than `fork`.
        ctx = mp_context or multiprocessing.get_context("spawn")

        self.manager = ctx.Manager()
        self.pids_for_job = self.manager.dict()
        self.timeouts: dict[str, asyncio.Task] = {}
        self.pool = ProcessPoolExecutor(
            max_workers=queue.concurrency,
            max_tasks_per_child=maximum_jobs_per_worker,
            initializer=self.on_initialize_worker,
            mp_context=ctx,
        )

    @classmethod
    def on_initialize_worker(cls):
        """
        This method is called in each worker process before it begins running
        jobs. It can be used to perform any necessary setup, such as loading
        NLTK datasets or calling ``django.setup()``.

        This isn't called once per job but once per worker process until
        :attr:`~ProcessExecutor.maximum_jobs_per_worker` is reached (if
        set). After that, the worker process is replaced with a new one.

        .. note::

            Care should be taken when overriding this method, as it is called
            within a separate process and may not have access to the same
            resources as the main process.
        """
        signal.signal(signal.SIGALRM, cls.job_signal_handler)
        signal.signal(signal.SIGUSR1, cls.job_signal_handler)

    async def push(self, job: QueuedJob) -> Future:
        future: Future = self.pool.submit(
            self.job_wrapper, job, self.pids_for_job
        )
        future.add_done_callback(
            functools.partial(
                self._on_job_completed, loop=asyncio.get_running_loop()
            )
        )
        self.jobs[future] = job
        time_limit = next(
            (
                limit.value
                for limit in job.limits
                if limit.type_ == Limit.Type.TIME
            ),
            None,
        )
        if time_limit is not None:
            self.timeouts[job.id] = asyncio.create_task(
                self._handle_timeout(job.id, time_limit)
            )

        return future

    async def _handle_timeout(self, job_id: str, time_limit: int):
        try:
            await asyncio.sleep(time_limit)
            pid = self.pids_for_job.get(job_id)
            if pid is not None:
                os.kill(pid, signal.SIGALRM)
        except asyncio.CancelledError:
            pass

    @classmethod
    def job_wrapper(cls, job: QueuedJob, pids_for_job):
        """
        This is the function that is actually started by the process pool
        executor. It's responsible for setting up necessary signals and limits,
        running the job, and returning the result.

        Subclasses can override this method to provide additional functionality
        or to change the way that jobs are run.

        .. note::

            Care should be taken when overriding this method, as it is called
            within a separate process and may not have access to the same
            resources as the main process.
        """
        cleanup: list[Callable] = []
        try:
            pids_for_job[job.id] = os.getpid()
            func, kwargs = cls.get_function_and_kwargs(job)
            if asyncio.iscoroutinefunction(func):
                raise ValueError(
                    f"Function {job.func!r} is an async function, which is not"
                    f" supported by the {cls.__name__!r} executor. Use the "
                    "AsyncExecutor instead."
                )

            for limit in job.limits:
                match limit.type_:
                    case Limit.Type.MEMORY:
                        previous_soft, _ = resource.getrlimit(
                            resource.RLIMIT_AS
                        )
                        resource.setrlimit(
                            resource.RLIMIT_AS, (limit.value, -1)
                        )
                        cleanup.append(
                            lambda: resource.setrlimit(
                                resource.RLIMIT_AS, (previous_soft, -1)
                            )
                        )

            func(**kwargs)
        finally:
            pids_for_job.pop(job.id)
            for clean in cleanup:
                clean()

        return job

    @staticmethod
    def job_signal_handler(signum: int, frame):
        """
        Handles signals sent to a running job process.

        Subclasses can override this method to provide additional functionality
        or to change the way that signals are handled.

        .. note::

            Care should be taken when overriding this method, as it is called
            within a separate process and may not have access to the same
            resources as the main process.
        """
        if signum == signal.SIGALRM:
            raise TimeoutError("Job timed out.")
        if signum == signal.SIGUSR1:
            raise CancelledError("Job was cancelled.")

    def _on_job_completed(
        self, future: Future, loop: asyncio.AbstractEventLoop
    ):
        job = self.jobs.pop(future)

        timeout_task = self.timeouts.pop(job.id, None)
        if timeout_task is not None:
            timeout_task.cancel()

        exc = future.exception()
        if exc is None:
            job = future.result()

        f = asyncio.run_coroutine_threadsafe(
            self.job_completed(job, exc),
            loop,
        )
        f.result()

    async def stop(self):
        for task in self.timeouts.values():
            task.cancel()

        self.pool.shutdown(cancel_futures=True)
        self.manager.shutdown()

    async def cancel(self, ref: Reference):
        """
        Make an attempt to cancel a running job.

        It's not guaranteed that the job will be cancelled, nor is it
        guaranteed that the job will be cancelled in a timely manner. For
        example if the job is running a long computation in a C extension,
        it may not be possible to interrupt it until it returns.

        :param ref: The reference to the job to cancel.
        """
        await super().cancel(ref)
        pid = self.pids_for_job.get(ref.identifier)
        if pid is not None:
            os.kill(pid, signal.SIGUSR1)



import asyncio
import inspect
from dataclasses import dataclass
from typing import Callable, Awaitable


@dataclass
class Event:
    name: str
    body: dict[str, any]


EventCallbackT = Callable[[Event], Awaitable[None]] | Callable[[Event], None]


class Hub:
    """
    An event hub for registering and emitting events.
    """

    def __init__(self):
        self._handlers: dict[str, list[EventCallbackT]] = {}
        self._waiters: dict[str, set[asyncio.Future]] = {}

    def on(self, event: str, f: EventCallbackT):
        """
        Register a handler for an event.
        """
        self._handlers.setdefault(event, []).append(f)

    def remove(self, event: str, handler: EventCallbackT):
        """
        Remove a handler from an event.
        """
        try:
            self._handlers.get(event, []).remove(handler)
        except ValueError:
            pass

    def clear(self):
        """
        Clear all handlers.
        """
        self._handlers = {}

        for event in self._waiters:
            for waiter in self._waiters[event]:
                if not waiter.done():
                    waiter.cancel()

        self._waiters = {}

    async def emit(self, event: str, body):
        """
        Emit an event with the given body.
        """
        result = Event(name=event, body=body)

        for handler in self._handlers.get(event, []):
            if inspect.iscoroutinefunction(handler):
                await handler(result)
            else:
                handler(result)

        # Notify any asyncio waiters that are waiting for this event, used
        # to implement wait_for().
        waiters = self._waiters.get(event, [])
        for waiter in waiters:
            if not waiter.done():
                waiter.set_result(result)

        self._waiters[event] = {w for w in waiters if not w.done()}

    async def wait_for(
        self, event: str | list[str], timeout: float | None = None
    ) -> list[Event]:
        """
        Wait for a specific event to occur.

        :param event: The name of the event to wait for, or a list of events.
        :param timeout: The maximum time to wait for the event (in seconds).
        """
        if isinstance(event, str):
            event = [event]

        if not event:
            raise ValueError("At least one event must be provided.")

        futures = []
        for e in event:
            future = asyncio.get_running_loop().create_future()
            future.add_done_callback(
                lambda _: self._waiters.get(e, set()).discard(future)
            )
            self._waiters.setdefault(e, set()).add(future)
            futures.append(future)

        done, pending = await asyncio.wait(
            futures,
            timeout=timeout,
            return_when=asyncio.FIRST_COMPLETED,
        )

        for task in pending:
            task.cancel()

        return [task.result() for task in done]



import json
from datetime import datetime, timezone

from psycopg import sql
from croniter import croniter

from chancy.plugin import Plugin, PluginScope
from chancy.worker import Worker
from chancy.app import Chancy
from chancy import Job


class Cron(Plugin):
    """
    Run jobs at specific times and intervals using cron-like syntax.

    The schedule is persistent and dynamic, being stored in the database. This
    allows for jobs to be scheduled and rescheduled without needing to restart
    the worker(s).

    If a scheduled job is already on the queue waiting to run, or currently
    running, the job will not be queued again and instead will wait until the
    next scheduled time.

    .. note::

        While the underlying library used to parse the cron syntax supports
        timezones, this plugin does not. All times are assumed to be in UTC.
        This is due to frequent issues that occur with timezones and daylight
        saving time changes that we simply don't want to support.

    Installation
    ------------

    This plugin requires extra dependencies to parse the cron syntax. You can
    install them using:

    .. code-block:: bash

        pip install chancy[cron]

    This plugin requires a database migration to create the table that stores
    the cron-like schedules.

    Usage
    -----

    To use the cron plugin, you need to add it to your Chancy application and
    then set up the schedule for the jobs you want to run:

    .. code-block:: python
        :caption: worker.py

        import asyncio
        from chancy import Chancy, Worker, Queue
        from chancy.plugins.pruner import Cron
        from chancy.plugins.leadership import Leadership

        def hello_world():
            print("hello_world")

        async with Chancy(
            dsn="postgresql://localhost/postgres",
            plugins=[
                Queue(name="default", concurrency=10),
                Leadership(),
                Cron(),
            ],
        ) as chancy:
            await chancy.migrate()
            await Cron.schedule(
                chancy,
                "*/2 * * * *",
                Job(
                    func="worker.hello_world",
                    unique_key="hello_world_cron",
                    queue="default",
                )
            )
            await Worker(chancy).start()

    :param poll_interval: The number of seconds between cron poll intervals.
    """

    def __init__(self, *, poll_interval: int = 60):
        super().__init__()
        self.poll_interval = poll_interval

    @classmethod
    def get_scope(cls) -> PluginScope:
        return PluginScope.WORKER

    async def run(self, worker: Worker, chancy: Chancy):
        table = sql.Identifier(f"{chancy.prefix}cron")

        while await self.sleep(self.poll_interval):
            async with chancy.pool.connection() as conn:
                # We need to find every row in the {prefix}_cron table where
                # the next_run time is less than or equal to the current time,
                # lock it, update the next_run time, and then push the job onto
                # the queue.
                now = datetime.now(tz=timezone.utc)
                async with conn.cursor() as cursor:
                    async with conn.transaction():
                        await cursor.execute(
                            sql.SQL(
                                """
                                SELECT
                                    unique_key,
                                    cron,
                                    job
                                FROM {table}
                                WHERE next_run <= %(now)s
                                FOR UPDATE SKIP LOCKED
                                """
                            ).format(table=table),
                            {"now": now},
                        )

                        for row in await cursor.fetchall():
                            unique_key, cron, job = row
                            # If we're using our built-in default queue, we
                            # can push this as part of our transaction.
                            await chancy.push_many_ex(
                                cursor,
                                [Job.unpack(job)],
                            )

                            chancy.log.debug(
                                f"Pushed scheduled cron job {unique_key!r}"
                            )

                            await cursor.execute(
                                sql.SQL(
                                    """
                                    UPDATE {table}
                                    SET
                                        next_run = %(next_run)s,
                                        last_run = %(last_run)s
                                    WHERE unique_key = %(unique_key)s
                                    """
                                ).format(table=table),
                                {
                                    "next_run": croniter(cron, now).get_next(
                                        datetime
                                    ),
                                    "last_run": now,
                                    "unique_key": unique_key,
                                },
                            )

    def migrate_key(self) -> str | None:
        return "cron"

    def migrate_package(self) -> str | None:
        return "chancy.plugins.cron.migrations"

    def api_plugin(self) -> str | None:
        return "chancy.plugins.cron.api.CronApiPlugin"

    @classmethod
    async def unschedule(cls, chancy: Chancy, *unique_keys: str):
        """
        Permanently unschedule one or more jobs from running.

        .. code-block:: python

            await Cron.unschedule(chancy, "hello_world_cron")

        :param chancy: The Chancy application.
        :param unique_keys: The unique keys of the jobs to unschedule.
        """
        async with chancy.pool.connection() as conn:
            async with conn.cursor() as cursor:
                async with conn.transaction():
                    await cursor.execute(
                        sql.SQL(
                            """
                            DELETE FROM {table}
                            WHERE unique_key = ANY(%(unique_keys)s)
                            """
                        ).format(table=sql.Identifier(f"{chancy.prefix}cron")),
                        {"unique_keys": unique_keys},
                    )

    @classmethod
    async def schedule(cls, chancy: Chancy, cron: str, *jobs: Job):
        """
        Schedule one or more jobs to run at specific times and intervals.

        All jobs that are scheduled with this feature *must* be using a
        :attr:`~chancy.job.Job.unique_key` to ensure that only one
        copy of the job is scheduled at a time. Scheduling a job with the same
        unique key as an existing job will update the existing job with the new
        schedule & job.

        For example, to run a function once every 2 minutes, we'd use:

        .. code-block:: python

            await Cron.schedule(
                chancy,
                "default",
                "*/2 * * * *",
                Job(func="worker.hello_world", unique_key="hello_world_cron")
            )

        :param chancy: The Chancy application.
        :param cron: A cron-like syntax string that describes when to run the
                     job.
        :param jobs: The jobs to run.
        """
        jobs = list(jobs)
        for job in jobs:
            if not job.unique_key:
                raise ValueError(
                    "Scheduling jobs for execution on a cron-like schedule"
                    " requires that each job has a unique_key set."
                )

        base = datetime.now(tz=timezone.utc)

        async with chancy.pool.connection() as conn:
            async with conn.cursor() as cursor:
                async with conn.transaction():
                    await cursor.executemany(
                        sql.SQL(
                            """
                            INSERT INTO {table} (
                                unique_key,
                                cron,
                                job,
                                next_run
                            )
                            VALUES (
                                %(unique_key)s,
                                %(cron)s,
                                %(job)s,
                                %(next_run)s
                            )
                            ON CONFLICT (unique_key) DO UPDATE SET
                                cron = %(cron)s,
                                job = %(job)s,
                                next_run = %(next_run)s
                            """
                        ).format(table=sql.Identifier(f"{chancy.prefix}cron")),
                        [
                            {
                                "unique_key": job.unique_key,
                                "cron": cron,
                                "job": json.dumps(job.pack()),
                                "next_run": croniter(cron, base).get_next(
                                    datetime
                                ),
                            }
                            for job in jobs
                        ],
                    )



from dataclasses import asdict

from psycopg.rows import dict_row
from starlette.responses import Response

from chancy import Job
from chancy.plugins.api import ApiPlugin
from chancy.utils import json_dumps


class CronApiPlugin(ApiPlugin):
    """
    Provides API endpoints for viewing cron jobs.
    """

    def name(self):
        return "cron"

    def routes(self):
        return [
            {
                "path": "/api/v1/crons",
                "endpoint": self.get_cron,
                "methods": ["GET"],
                "name": "get_cron",
            },
        ]

    @staticmethod
    async def get_cron(request, *, chancy, worker):
        """
        Get all known cron jobs.
        """
        async with chancy.pool.connection() as conn:
            async with conn.cursor(row_factory=dict_row) as cursor:
                await cursor.execute(
                    """
                    SELECT * FROM {table}
                    """.format(table=f"{chancy.prefix}cron")
                )

                results = await cursor.fetchall()

                return Response(
                    json_dumps(
                        [
                            {
                                **result,
                                "job": asdict(Job.unpack(result["job"])),
                            }
                            for result in results
                        ]
                    ),
                    media_type="application/json",
                )



import asyncio
from asyncio import CancelledError

from chancy import Reference
from chancy.executors.base import Executor
from chancy.job import QueuedJob, Limit


class AsyncExecutor(Executor):
    """
    An Executor which uses asyncio to run its jobs in the main event loop.

    This executor is useful for running large numbers of IO-bound jobs, as it
    can run many jobs concurrently without blocking the main event loop and
    without the high overhead of new processes or threads. However, it is not
    suitable for CPU-bound jobs, as it will block the main event loop and
    prevent other jobs & queues from running.

    To use this executor, simply pass the import path to this class in the
    ``executor`` field of your queue configuration or use the
    :class:`~chancy.app.Chancy.Executor` shortcut:

    .. code-block:: python

        async with Chancy(dsn="postgresql://localhost/postgres") as chancy:
            await chancy.declare(
                Queue(
                    name="default",
                    executor=Chancy.Executor.Async,
                )
            )
    """

    def __init__(self, worker, queue):
        super().__init__(worker, queue)
        self.jobs: dict[asyncio.Task, QueuedJob] = {}

    async def push(self, job: QueuedJob):
        task = asyncio.create_task(self._job_wrapper(job))
        self.jobs[task] = job
        task.add_done_callback(self.jobs.pop)

    def __len__(self):
        return len(self.jobs)

    async def _job_wrapper(self, job: QueuedJob):
        try:
            func, kwargs = Executor.get_function_and_kwargs(job)
            if not asyncio.iscoroutinefunction(func):
                raise ValueError(
                    f"Function {job.func!r} is not an async function, which is"
                    f" required for the AsyncExecutor. Please use the"
                    f" ThreadedExecutor or ProcessExecutor instead."
                )

            timeout = next(
                (
                    limit.value
                    for limit in job.limits
                    if limit.type_ == Limit.Type.TIME
                ),
                None,
            )

            async with asyncio.timeout(timeout):
                await func(**kwargs)
            await self.job_completed(job)
        except (Exception, CancelledError) as exc:
            await self.job_completed(job, exc)

    async def cancel(self, ref: Reference):
        for task, job in self.jobs.items():
            if job.id == ref.identifier:
                task.cancel()
                return

    async def stop(self):
        """
        Stop the executor, giving it a chance to clean up any resources it
        may have allocated to running jobs.
        """
        for task in self.jobs:
            task.cancel()
        await asyncio.gather(*self.jobs)



import re
import asyncio
import json
import time
import uuid
import socket
import signal
import platform

from psycopg import sql
from psycopg import AsyncConnection
from psycopg.rows import dict_row
from psycopg.types.json import Json

from chancy.app import Chancy
from chancy.executors.base import Executor
from chancy.hub import Hub, Event
from chancy.queue import Queue
from chancy.utils import TaskManager, import_string
from chancy.job import QueuedJob, Reference


class Worker:
    """
    The Worker is responsible for polling queues for new jobs, running any
    configured plugins, and internal management such as heartbeats.

    Starting a worker is simple:

    .. code-block:: python

        async with Chancy(
            dsn="postgresql://localhost/postgres",
        ) as chancy:
            await chancy.migrate()
            async with Worker(chancy) as worker:
                await worker.wait_until_complete()

    Worker Tags
    -----------

    Worker tags control which workers run which queues. A tag is a string
    (typically in the format ``key=value``) assigned to a worker. Declare
    a queue with a list of tags, and only workers with matching tags will
    run that queue.

    Every worker automatically gets some tags, like ``hostname``, ``worker_id``,
    ``python``, ``arch``, ``os``, and ``*`` (wildcard). You can also add custom
    tags when creating a worker.

    .. code-block:: python

        async with Worker(chancy, tags={"has=gpu", "has=large-disk"}) as worker:
            await worker.wait_until_complete()

    You could then assign a queue to only run on workers with the ``has=gpu`` tag:

    .. code-block:: python

        await chancy.declare(
            Queue(name="default", tags={"has=gpu"}, concurrency=10),
            upsert=True  # Replaces existing settings
        )

    Tags are regexes, allowing for flexible matching:

    .. code-block:: python

        Queue(name="default", tags={r"python=3\\.11\\.[0-9]+"}, concurrency=10)

    :param chancy: The Chancy application that the worker is associated with.
    :param worker_id: The ID of the worker, which must be globally unique. If
                      not provided, a random UUID will be generated.
    :param heartbeat_poll_interval: The number of seconds between heartbeat
                                    poll intervals.
    :param heartbeat_timeout: The number of seconds before a worker is
                              considered to have lost connection to the
                              cluster.
    :param queue_change_poll_interval: The number of seconds between checks for
                                       changes to the queues that the worker
                                       should be processing.
    :param shutdown_timeout: The number of seconds to wait for a clean shutdown
                                before forcing the worker to stop.
    :param tags: Extra tags to associate with the worker.
    :param register_signal_handlers: Whether to register signal handlers.
    """

    def __init__(
        self,
        chancy: Chancy,
        *,
        worker_id: str | None = None,
        heartbeat_poll_interval: int = 30,
        heartbeat_timeout: int = 90,
        queue_change_poll_interval: int = 10,
        shutdown_timeout: int = 30,
        tags: set[str] | None = None,
        register_signal_handlers: bool = True,
    ):
        #: The Chancy application that the worker is associated with.
        self.chancy = chancy
        #: The ID of the worker, which must be globally unique.
        self.worker_id = worker_id or str(uuid.uuid4())
        #: The number of seconds between heartbeat poll intervals.
        self.heartbeat_poll_interval = heartbeat_poll_interval
        #: The number of seconds before a worker is considered to have lost
        #: connection to the cluster.
        self.heartbeat_timeout = heartbeat_timeout
        #: The number of seconds to wait for a clean shutdown before forcing
        #: the worker to stop.
        self.shutdown_timeout = shutdown_timeout
        #: The number of seconds between sending outgoing updates to the
        #: database.
        self.send_outgoing_interval = 1
        #: How frequently to check for changes to the queues that this worker
        #: should be processing.
        self.queue_change_poll_interval = queue_change_poll_interval
        #: An event hub with cluster and local telemetry, useful for plugins.
        self.hub = Hub()
        #: Whether to register signal handlers on startup of the worker.
        self.register_signal_handlers = register_signal_handlers
        # Extra tags to associate with the worker.
        self._extra_tags = tags or set()

        #: The task manager for the worker, responsible for managing all the
        #: worker's internal asyncio tasks.
        self.manager = TaskManager()

        #: A queue of updates waiting to be sent to the database.
        self.outgoing: asyncio.Queue[QueuedJob] = asyncio.Queue()

        #: An event that is set when the worker is the leader.
        #: This functionality is not enabled by default - a leadership plugin
        #: must be used to enable this event.
        self.is_leader = asyncio.Event()
        #: An event that is set when the worker is shutting down due to
        #: receiving a signal.
        self.shutdown_event = asyncio.Event()

        # Set once the notifications listener is ready.
        self._notifications_ready_event = asyncio.Event()
        # The queues that the worker is currently processing.
        self._queues: dict[str, Queue] = {}
        # The executors that the worker is currently using.
        self._executors: dict[str, Executor] = {}

    async def start(self):
        """
        Start the worker and begin processing jobs.

        Unlike `run`, this method does not wait for the worker to complete and
        returns immediately.
        """
        self.hub.on("job.cancelled", self._handle_cancellation)

        for plugin in self.chancy.plugins:
            self.manager.add(
                plugin.run(self, self.chancy),
                name=plugin.__class__.__name__,
            )
            self.chancy.log.info(
                f"Started plugin {plugin.__class__.__name__!r}"
            )

        self.manager.add(self._maintain_queues(), name="queues")
        self.manager.add(self._maintain_updates(), name="updates")
        self.manager.add(self._maintain_heartbeat(), name="heartbeat")

        # pgbouncer and other connection pools may not support LISTEN/NOTIFY
        # properly, so this needs to remain an optional (but default) feature.
        if self.chancy.notifications:
            self.manager.add(
                self._maintain_notifications(), name="notifications"
            )
            await self._notifications_ready_event.wait()

        if self.register_signal_handlers:
            loop: asyncio.AbstractEventLoop = asyncio.get_running_loop()
            for sig in {signal.SIGTERM, signal.SIGINT}:
                loop.add_signal_handler(
                    sig,
                    lambda: asyncio.create_task(self.on_signal(sig)),
                )

    async def wait_until_complete(self):
        """
        Wait until the worker is stopped.
        """
        await self.manager.wait_until_complete()

    async def _maintain_queues(self):
        """
        Maintain the queues that the worker is processing.

        This will periodically check the database for any changes to the queues
        that the worker should be processing, and update the worker's queues
        accordingly.
        """
        while True:
            self.chancy.log.debug("Polling for queue changes.")
            async with self.chancy.pool.connection() as conn:
                async with conn.cursor(row_factory=dict_row) as cursor:
                    tags = self.worker_tags()
                    await cursor.execute(
                        sql.SQL(
                            """
                            SELECT 
                                name,
                                concurrency,
                                tags,
                                state,
                                executor,
                                executor_options,
                                polling_interval,
                                rate_limit,
                                rate_limit_window
                            FROM {queues}
                        """
                        ).format(
                            queues=sql.Identifier(f"{self.chancy.prefix}queues")
                        ),
                    )
                    # Tags in the database is a list of regexes, while the tags
                    # in the worker are a set of strings. We need to filter the
                    # queues based on the worker's tags.
                    db_queues = {
                        q["name"]: Queue.unpack(q)
                        for q in await cursor.fetchall()
                        if any(
                            re.match(t, tag) for tag in tags for t in q["tags"]
                        )
                    }

            for queue_name in list(self._queues.keys()):
                if queue_name not in db_queues:
                    del self._queues[queue_name]
                    self.chancy.log.info(f"Removed queue {queue_name}")

            for queue_name, queue in db_queues.items():
                if queue_name not in self._queues:
                    self._queues[queue_name] = queue
                    self.manager.add(
                        self._maintain_queue(queue),
                        name=f"queue_{queue_name}",
                    )
                    self.chancy.log.info(
                        f"Adding queue {queue_name!r} to worker using executor"
                        f" {queue.executor!r}."
                    )
                else:
                    if self._queues[queue_name] != queue:
                        self._queues[queue_name] = queue
                        self.chancy.log.info(
                            f"Updated configuration for the queue"
                            f" {queue_name!r}."
                        )

            try:
                await self.hub.wait_for(
                    "queue.declared",
                    timeout=self.queue_change_poll_interval,
                )
            except asyncio.TimeoutError:
                pass

    async def _maintain_queue(self, queue: Queue):
        """
        Maintain a single queue.

        Responsible for polling a queue for new jobs and pushing them to the
        executor.

        :param queue: The queue to maintain.
        """
        cls = import_string(queue.executor)

        async with cls(self, queue, **queue.executor_options) as executor:
            self._executors[queue.name] = executor

            while True:
                try:
                    queue = self._queues[queue.name]
                except KeyError:
                    self.chancy.log.info(
                        f"Queue {queue.name!r} no longer exists, stopping its"
                        f" maintenance loop."
                    )
                    self._executors.pop(queue.name)
                    return

                maximum_jobs_to_poll = queue.concurrency - len(executor)
                if maximum_jobs_to_poll <= 0:
                    await asyncio.sleep(queue.polling_interval)
                    continue

                if queue.state == Queue.State.PAUSED:
                    self.chancy.log.debug(
                        f"Queue {queue.name!r} is paused, skipping polling."
                    )
                    await asyncio.sleep(queue.polling_interval)
                    continue

                async with self.chancy.pool.connection() as conn:
                    jobs = await self.fetch_jobs(
                        queue, conn, up_to=maximum_jobs_to_poll
                    )

                for job in jobs:
                    self.chancy.log.info(
                        f"Pulled {job.id!r} ({job.func!r}) for queue {job.queue!r}"
                    )
                    await executor.push(job)

                await asyncio.sleep(queue.polling_interval)

    async def _maintain_heartbeat(self):
        """
        Announces the worker to the cluster, and maintains a periodic heartbeat
        to ensure that the worker is still alive.
        """
        while True:
            async with self.chancy.pool.connection() as conn:
                async with conn.transaction():
                    self.chancy.log.debug("Announcing worker to the cluster.")
                    await self.announce_worker(conn)
            await asyncio.sleep(self.heartbeat_poll_interval)

    async def _maintain_notifications(self):
        """
        Listen for notifications from the database.

        Improves the reactivity of a worker by allowing it to almost immediately
        react to cluster events using Postgres's LISTEN/NOTIFY feature.

        .. note::

            This feature utilizes a permanent connection to the database
            separate from the shared connection pool and is not counted against
            the pool's connection limit.
        """
        connection = await AsyncConnection.connect(
            self.chancy.dsn, autocommit=True
        )
        await connection.execute(
            sql.SQL("LISTEN {channel};").format(
                channel=sql.Identifier(f"{self.chancy.prefix}events")
            )
        )
        self.chancy.log.info("Started listening for realtime notifications.")
        self._notifications_ready_event.set()
        async for notification in connection.notifies():
            j = json.loads(notification.payload)
            await self.hub.emit(j.pop("t"), j)

    async def _maintain_updates(self):
        """
        Process updates to job instances.

        We maintain a queue of updates to job instances, and process them in
        batches to significantly reduce the number of overall transactions that
        need to be made, at the cost of potentially losing some updates if the
        worker is stopped unexpectedly. The frequency of these updates can be
        controlled by setting the `send_outgoing_interval` attribute on the
        worker.
        """
        while True:
            if self.outgoing.empty():
                await asyncio.sleep(self.send_outgoing_interval)
                continue

            pending_updates = []
            while len(pending_updates) < 1000:
                try:
                    pending_updates.append(self.outgoing.get_nowait())
                except asyncio.QueueEmpty:
                    break

            self.chancy.log.debug(
                f"Processing {len(pending_updates)} outgoing updates."
            )

            async with self.chancy.pool.connection() as conn:
                async with conn.cursor() as cursor:
                    async with conn.transaction():
                        try:
                            await cursor.executemany(
                                sql.SQL(
                                    """
                                    UPDATE
                                        {jobs}
                                    SET
                                        state = %(state)s,
                                        started_at = %(started_at)s,
                                        completed_at = %(completed_at)s,
                                        attempts = %(attempts)s,
                                        errors = %(errors)s,
                                        meta = %(meta)s,
                                        max_attempts = %(max_attempts)s
                                    WHERE
                                        id = %(id)s
                                    """
                                ).format(
                                    jobs=sql.Identifier(
                                        f"{self.chancy.prefix}jobs"
                                    )
                                ),
                                [
                                    {
                                        "id": update.id,
                                        "state": update.state.value,
                                        "started_at": update.started_at,
                                        "completed_at": update.completed_at,
                                        "attempts": update.attempts,
                                        "errors": Json(update.errors),
                                        "meta": Json(update.meta),
                                        "max_attempts": update.max_attempts,
                                    }
                                    for update in pending_updates
                                ],
                            )
                        except Exception:
                            # If we were unable to apply the updates, we should
                            # re-queue them for the next poll.
                            self.chancy.log.exception(
                                "Failed to apply updates to job instances."
                            )
                            for update in pending_updates:
                                await self.outgoing.put(update)
                            raise

            await asyncio.sleep(self.send_outgoing_interval)

    async def announce_worker(self, conn: AsyncConnection):
        """
        Announce the worker to the cluster.

        This will insert the worker into the workers table, or update the
        `last_seen` timestamp if the worker is already present.

        :param conn: The connection to use for the announcement.
        """
        async with conn.cursor() as cur:
            async with conn.transaction():
                await cur.execute(
                    sql.SQL(
                        """
                        DELETE FROM {workers}
                        WHERE expires_at < NOW()
                        """
                    ).format(
                        workers=sql.Identifier(f"{self.chancy.prefix}workers")
                    )
                )
                await cur.execute(
                    sql.SQL(
                        """
                        INSERT INTO {workers}
                            (worker_id, last_seen, expires_at, tags, queues)
                        VALUES (
                            %(worker_id)s,
                            NOW(),
                            NOW() + INTERVAL {timeout},
                            %(tags)s,
                            %(queues)s
                        )
                        ON CONFLICT (worker_id) DO UPDATE
                            SET last_seen = NOW(),
                                expires_at = NOW() + INTERVAL {timeout},
                                tags = EXCLUDED.tags,
                                queues = EXCLUDED.queues
                        """
                    ).format(
                        workers=sql.Identifier(f"{self.chancy.prefix}workers"),
                        timeout=sql.Literal(
                            f"{self.heartbeat_timeout} seconds"
                        ),
                    ),
                    {
                        "worker_id": self.worker_id,
                        "tags": list(self.worker_tags()),
                        "queues": list(self._queues.keys()),
                    },
                )
                await self.chancy.notify(
                    cur, "worker.announced", {"worker_id": self.worker_id}
                )

    def worker_tags(self) -> set[str]:
        """
        Return the tags associated with the worker.

        Tags are used to limit which Queues get assigned to a worker, and
        may also be used by some plugins.

        :return: A set of tags associated with the worker.
        """
        return {
            "*",
            f"hostname={socket.gethostname()}",
            f"worker_id={self.worker_id}",
            f"python={platform.python_version()}",
            f"os={platform.system()}",
            f"arch={platform.machine()}",
            *self._extra_tags,
        }

    async def queue_update(self, update: QueuedJob):
        """
        Enqueue an update to a job instance.

        This method will queue an update to a job instance to be processed in
        periodic batches, reducing the number of transactions that need to be
        made. You can control the frequency of these updates by setting the
        `send_outgoing_interval` attribute on the worker.

        :param update: The job instance to update.
        """
        await self.outgoing.put(update)

    async def fetch_jobs(
        self,
        queue: Queue,
        conn: AsyncConnection,
        *,
        up_to: int = 1,
    ) -> list[QueuedJob]:
        """
        Fetch jobs from the queue for processing.

        This method will fetch up to `up_to` jobs from the queue, mark them as
        running, and return them as a list of `QueuedJob` objects. If no jobs
        are available, an empty list will be returned.

        It's safe to call this method concurrently, as the jobs will be locked
        for the duration of the transaction.

        :param queue: The queue to fetch jobs from.
        :param conn: The database connection to use.
        :param up_to: The maximum number of jobs to fetch.
        """
        jobs_table = sql.Identifier(f"{self.chancy.prefix}jobs")
        rate_limits_table = sql.Identifier(
            f"{self.chancy.prefix}queue_rate_limits"
        )

        async with conn.cursor(row_factory=dict_row) as cursor:
            async with conn.transaction():
                # If the queue is configured to use a rate limit, we need to
                # check if there's any remaining capacity in the current
                # window.
                if queue.rate_limit:
                    now = int(time.time())
                    window_start = now - (now % queue.rate_limit_window)

                    await cursor.execute(
                        sql.SQL(
                            """
                            INSERT INTO {rate_limits_table} (
                                queue,
                                window_start,
                                count
                            )
                            VALUES (%s, %s, 0)
                            ON CONFLICT (queue) DO UPDATE
                            SET 
                                count = CASE
                                    WHEN {rate_limits_table}.window_start
                                        = EXCLUDED.window_start
                                            THEN {rate_limits_table}.count
                                    ELSE 0
                                END,
                                window_start = EXCLUDED.window_start
                            RETURNING count
                            """
                        ).format(rate_limits_table=rate_limits_table),
                        (queue.name, window_start),
                    )

                    result = await cursor.fetchone()
                    current_count = result["count"]

                    # If we've hit the rate limit, return early with no jobs
                    # fetched.
                    if current_count >= queue.rate_limit:
                        return []

                    # Adjust up_to based on remaining rate limit
                    up_to = min(up_to, queue.rate_limit - current_count)

                await cursor.execute(
                    sql.SQL(
                        """
                        WITH selected_jobs AS (
                            SELECT
                                id
                            FROM
                                {jobs}
                            WHERE
                                queue = %(queue)s
                            AND
                                (state = 'pending' OR state = 'retrying')
                            AND
                                attempts < max_attempts
                            AND
                                (scheduled_at IS NULL OR scheduled_at <= NOW())
                            ORDER BY
                                priority DESC,
                                id DESC
                            LIMIT
                                %(maximum_jobs_to_fetch)s
                            FOR UPDATE OF {jobs} SKIP LOCKED
                        )
                        UPDATE
                            {jobs}
                        SET
                            started_at = NOW(),
                            state = 'running',
                            taken_by = %(worker_id)s
                        FROM
                            selected_jobs
                        WHERE
                            {jobs}.id = selected_jobs.id
                        RETURNING {jobs}.*
                        """
                    ).format(
                        jobs=jobs_table,
                    ),
                    {
                        "queue": queue.name,
                        "maximum_jobs_to_fetch": up_to,
                        "worker_id": self.worker_id,
                    },
                )

                records = await cursor.fetchall()

                # If a rate limit is configured, and we ended up fetching jobs,
                # we need to increment the rate limit counter.
                if queue.rate_limit and records:
                    await cursor.execute(
                        sql.SQL(
                            """
                            UPDATE {rate_limits_table}
                            SET count = count + %s
                            WHERE queue = %s
                            """
                        ).format(rate_limits_table=rate_limits_table),
                        (len(records), queue.name),
                    )

                return [QueuedJob.unpack(record) for record in records]

    async def on_signal(self, signum: int):
        """
        Called when a signal is received.

        By default, handles SIGTERM and SIGINT by starting worker shutdown. If
        the signal is received again, it will force a shutdown.
        """
        if signum in (signal.SIGTERM, signal.SIGINT):
            # If the worker is already shutting down, we force an immediate
            # shutdown.
            if self.shutdown_event.is_set():
                self.chancy.log.warning(
                    "Received signal again, forcing shutdown."
                )
                asyncio.get_running_loop().stop()

            self.shutdown_event.set()
            await self.manager.cancel_all()

    async def stop(self):
        """
        Stop the worker.

        Attempts to stop the worker gracefully, sending a CancelledError to all
        running tasks and waiting up to `shutdown_timeout` seconds for them to
        complete before returning.
        """
        async with asyncio.timeout(self.shutdown_timeout):
            await self.manager.cancel_all()

    async def _handle_cancellation(self, event: Event):
        self.chancy.log.info(
            f"Received cancellation request for job {event.body['j']!r}"
        )
        for executor in self._executors.values():
            await executor.cancel(Reference(event.body["j"]))

    def __repr__(self):
        return f"<Worker({self.worker_id!r})>"

    def __iter__(self):
        return iter(self._queues.values())

    def __getitem__(self, key):
        return self._queues[key]

    def __contains__(self, key):
        return key in self._queues

    def __len__(self):
        return len(self._queues)

    async def __aenter__(self):
        await self.start()
        return self

    async def __aexit__(self, *exc):
        await self.stop()
